{*> Ver: Beta V1.2b8 *********      History      ***************************\

Beta V1.0b0		04/20/2001		Released
Beta V1.0b1		04/30/2001		Upstream notification added.
                                                Standartized notification / Up/Downstream
                                                Component iteraction functionality added

Beta V1.0b2  		04/31/2001		Improved editor.
                                                Added stream type registration.

Beta V1.0b3		05/02/2001		IgnoreUpstream has been added for the sinks.
Beta V1.0b4		05/03/2001		TOWPinList has been added.
Beta V1.0b5		05/04/2001		Remove notification has been added.
Beta V1.0b6		05/06/2001		TOWPinListOwner has been added.
Beta V1.0b7		05/07/2001		FMinPins and FMaxPins have been added to TOWPinListOwner.
                                                Fixed remove notifycation on destroy.

Beta V1.0b8		05/09/2001		TOWNotificationState has been added.
Beta V1.0b9		05/11/2001		Partial Delphi 6 support has been added.
                                                Improved TOWPinListPropertyEditor.
                                                Pin Categories ave been added.
                                                TOWNotificationResults has been adde in order to support dynamic balansing in the future.

Beta V1.1b0		06/04/2001		Multiform support. Dual link information.
                                                Improved ObjectInspector refresh.
                                                Various fixes.

Beta V1.2b0		06/18/2001		Dynamic order optimization.
                                                Sink edit property editor.

Beta V1.2b1		06/25/2001		Some fixes in the multiform support.
Beta V1.2b2		07/11/2001		RequestConnection has been replaced by Connect
                                                and now is part of TOWPin.

Beta V1.2b3		07/11/2001		Improved TOWPinList support and property editors.
Beta V1.2b5		07/19/2001		Disconnect procedure has been added.
Beta V1.2b6		08/01/2001		Minor fixes for the graphic editor support.
Beta V1.2b7		08/06/2001		Added pin add/remove/change notification for experts.
Beta V1.2b8		08/13/2001		Edded function dependancy check.

Legal issues: Copyright (C) 2001 by Boian Mitov
              mitov@mitov.com
              www.mitov.com
              www.openwire.org

              This software is provided 'as-is', without any express or
              implied warranty.  In no event will the author be held liable
              for any  damages arising from the use of this software.

              Permission is granted to anyone to use this software for any
              purpose, including commercial applications, and to alter it
              and redistribute it freely, subject to the following
              restrictions:

              1. The origin of this software must not be misrepresented,
                 you must not claim that you wrote the original software.
                 If you use this software in a product, an acknowledgment
                 in the product documentation would be appreciated but is
                 not required.

              2. Altered source versions must be plainly marked as such, and
                 must not be misrepresented as being the original software.

              3. This notice may not be removed or altered from any source
                 distribution.

Acknowledgement: Special thanks to Atanas Stoyanov for his idea of using the
                 Interfaces instead oh plain GUIDs.

TODO :
                Restricting and renegotiating handshaking with different type of data.
                Graph editor.

\***************************************************************************}
unit OpenWirePins;

interface

uses classes, windows, typinfo, Sysutils;

var  PinsNeedRefresh : Boolean;

                                
type
  PGUID = ^TGUID;        

  IOWStream = interface
    ['{2BFF1BE1-1698-4CFA-A427-9E0801C5B357}']
  end;

//---------------------------------------------------------------------------
TOWNotificationState   = set of ( nsNewLink, nsLastIteration );
TOWNotificationResults = set of ( nrDataChanged );
//---------------------------------------------------------------------------
type TOWSubmitFunction = function ( Handler : IOWStream; Operation : Integer; Data : Pointer; State : TOWNotificationState ) : TOWNotificationResults of object;
//---------------------------------------------------------------------------
TOWSinkPin = class;
//---------------------------------------------------------------------------
TPinTypeEntry=record
  ID             : TGUID;
  SubmitFunction : TOWSubmitFunction;
end;

TPinType=class( TObject )
protected
  Types : array of TPinTypeEntry;

public

public
  procedure AddType( ID : TGUID; SubmitFunction : TOWSubmitFunction );

  function  GetID( Index : Integer )             : TGUID;
  function  GetSubmitFunc( Index : Integer )     : TOWSubmitFunction;
  function  GetLength()                          : Integer;

public
  property  ID [ Index : Integer ] : TGUID read GetID;
  property  SubmitFunc [ Index : Integer ] : TOWSubmitFunction read GetSubmitFunc;
  property  Count : Integer read GetLength;

end;

//---------------------------------------------------------------------------
TOWPin = class(TPersistent, IOWStream)

protected
  FOwner                : TComponent;
  StreamType            : TPinType;
  FDeleteNotifyes       : TList;
  InSending             : Boolean;

public
  constructor Create(AOwner: TComponent);
  destructor Destroy; override;

public
  procedure AfterConstruction(); override;
  procedure BeforeDestruction(); override;

protected
  function  QueryInterface(const IID: TGUID; out Obj): HResult; virtual; stdcall;
  function  _AddRef: Integer; stdcall;
  function  _Release: Integer; stdcall;

protected
  procedure AddType( ID : TGUID );  overload;
  procedure AddType( ID : TGUID; SubmitFunction : TOWSubmitFunction ); overload;

protected
  procedure AddDeleteNotify( ADeleteNotifyPin : TOWPin );
//  procedure RemoveDeleteNotify( ADeleteNotifyPin : TOWPin );

protected
  procedure PinDeletedNotify( ADeletedPin : TOWPin ); virtual;

protected
  function  CheckInterfaces ( const OtherPin : TOWPin ) : Boolean; virtual;
  function  TypeCheckInterfaces ( const OtherPin : TOWPin ) : Boolean; virtual;
  function  GetSubmitFunction ( const OtherPin : TOWPin ) : TOWSubmitFunction;
  function  GetLinkName ( const OtherPin : TOWPin ) : String;

public
  function  Connect( OtherPin : TOWPin ) : Boolean; virtual; abstract;
  procedure Disconnect(); virtual; abstract;

public
  function  Notify( Operation : Integer; Data : Pointer ) : TOWNotificationResults; virtual; abstract;

public
  function  DefaultSubmit( Handler : IOWStream; Operation : Integer; Data : Pointer; State : TOWNotificationState ) : TOWNotificationResults; virtual;

public
  function  IsCompatible( const OtherPin : TOWPin ) : Boolean; virtual; abstract;
  // Wether or not the OtherPin is compatible with the entire data chain of already connected slots.
  function  CanConnectTo( const OtherPin : TOWPin ) : Boolean; virtual;
  function  DependsOn( const OtherPin : TOWPin ) : Boolean; virtual; abstract;
  function  IsConnected() : Boolean; virtual; abstract;

protected
  function  IsCompatibleChain( const OtherPin : TOWPin ) : Boolean; virtual;

public
  property  Owner : TComponent read FOwner;

end;
//---------------------------------------------------------------------------
EOWException = class(Exception)
end;

//---------------------------------------------------------------------------
TPinEntry=record
  SinkPin               : TOWSinkPin;
  SubmitFunction        : TOWSubmitFunction;
  ModificationLevel     : Integer; // Indicates how often this pin is source of data modification.
end;

//---------------------------------------------------------------------------

TOWSourcePin = class(TOWPin)
protected
  SinkPins              : TList;
  FFunctionSources      : TList;
  FDataTypeSources      : TList;
  InDependOn            : Boolean;

protected
  FLinkStorage          : TStringList;

public
  CurrentEditorPtr      : ^TOWSourcePin;

protected
  procedure IntConnect( SinkPin : TOWSinkPin );
  procedure IntDisconnect( SinkPin : TOWSinkPin );

public
  function  GetSinkCountAll () : Integer;  //                        virtual; // const;

public
  function  IsLinkedTo ( SinkName : String ) : Boolean; overload;
  
protected
  function  GetSinkCount() : Integer;                           virtual; // const;
  function  GetSink( Index : Integer ) : TOWSinkPin;            virtual;
  procedure SinkChangedData( Index : Integer );
  procedure ReorderChangedData();
  function  GetSubmit( Index : Integer ) : TOWSubmitFunction;   virtual;

  function  GetFunctionSourceCount () : Integer; // const;
  function  GetFunctionSources ( Index : Integer ) : TOWSinkPin;

  function  GetDataTypeSourceCount () : Integer; // const;
  function  GetDataTypeSources ( Index : Integer ) : TOWSinkPin;

  procedure IntAddFunctionSource ( FunctionSourcePin : TOWSinkPin );
  procedure IntRemoveFunctionSource ( FunctionSourcePin : TOWSinkPin );

  procedure IntAddDataTypeSource ( DataTypeSourcePin : TOWSinkPin );
  procedure IntRemoveDataTypeSource ( DataTypeSourcePin : TOWSinkPin );

public
  function  GetLinkStr( Item : Integer; SaveFormat : Boolean ) : String;
  procedure ClearPending();
  
protected
  procedure PinDeletedNotify( ADeletedPin : TOWPin ); override;
  function  GetLinkStrInt( Offset : String ) : String;

protected
  function  ChainCheckInterfaces ( ASinkPin : TOWSinkPin ) : Boolean;

protected
  procedure NewConnection( ASinkPin : TOWSinkPin; ASubmitFunction : TOWSubmitFunction ); virtual;
  
protected
  procedure ReadStrData ( Reader : TReader );
  procedure WriteStrData( Writer : TWriter );

  function  DoWrite() : Boolean;

  procedure DefineProperties(Filer: TFiler); override;

protected
  function  DoTryLink( Root : TComponent; SaveForm : Boolean ): Boolean;
  procedure AddToStorage( Ident : String );

public
  function  TryLinkTo ( Root : TComponent; SinkName : String; TryLater : Boolean; SaveForm : Boolean ) : Boolean;
  
protected
  procedure TryLinkToList ( Root : TComponent; SinkNames : TStringList );

public
  function  Connect( OtherPin : TOWPin ) : Boolean; override;
  function  IsCompatible ( const OtherPin : TOWPin ) : Boolean; override;
//  function  IsCompatibleChain ( const OtherPin : TOWPin ) : Boolean; override;
  function  CanConnectTo ( const OtherPin : TOWPin ) : Boolean; override;
  function  DependsOn( const OtherPin : TOWPin ) : Boolean; override;
  function  IsConnected() : Boolean; override;
//  function  GetStreamType () : TPinType; virtual; // const
//  function  GetChainStreamType () : TPinType; virtual;//  const;

  function  IsLinkedTo( OtherPin : TOWPin ) : Boolean;  overload;//const;
  procedure Disconnect(); override;

  procedure AddFunctionSource ( FunctionSourcePin : TOWSinkPin );
  procedure RemoveFunctionSource ( FunctionSourcePin : TOWSinkPin );

  procedure AddDataTypeSource ( DataTypeSourcePin : TOWSinkPin );
  procedure RemoveDataTypeSource ( DataTypeSourcePin : TOWSinkPin );

public
  function  Notify( Operation : Integer; Data : Pointer ) : TOWNotificationResults; override;

public
  constructor Create(AOwner: TComponent);
  destructor  Destroy; override;

public
  property SinkCount : Integer read GetSinkCount;
  property Sinks [ Index : Integer  ] : TOWSinkPin read GetSink;

  // Describes the functions sources for the output within the component.
  property FunctionSourceCount : Integer read GetFunctionSourceCount;
  property FunctionSources [ Index : Integer ] : TOWSinkPin read GetFunctionSources;

  // Describes the data type sources for the output within the component( The output should be of the same type as them ).
  property DataTypeSourceCount : Integer read GetDataTypeSourceCount;
  property DataTypeSources [ Index : Integer ] : TOWSinkPin read GetDataTypeSources;

end;

//---------------------------------------------------------------------------

TOWSinkPin = class(TOWPin)
public
  CurrentEditorPtr      : ^TOWSinkPin;

protected
  FSourcePin            : TOWSourcePin;
  SubmitFunction        : TOWSubmitFunction;

  FIgnoreUpstream       : Boolean;

  FDownStreamLinkName   : String;
  FUpStreamLinkName     : String;

  FFunctionDependants   : TList;
  FDataTypeDependants   : TList;

protected
  FInTryLink     : Boolean;
  FLinkStorage   : String;

protected
  procedure PinDeletedNotify( ADeletedPin : TOWPin ); override;

public
  function  TryLinkTo ( Root : TComponent; Soutce : String; TryLater : Boolean; SaveForm : Boolean ) : Boolean;

protected
  procedure IntAddFunctionDependance ( FunctionDependancePin : TOWSourcePin );
  procedure IntRemoveFunctionDependance ( FunctionDependancePin : TOWSourcePin );
  procedure IntAddDataTypeDependance ( DataTypeDependancePin : TOWSourcePin );
  procedure IntRemoveDataTypeDependance ( DataTypeDependancePin : TOWSourcePin );

protected
//  function  CheckInterfaces ( const OtherPin : TOWPin ) : Boolean; override;

  procedure SetSourcePin ( SourcePin : TOWSourcePin );
  function  GetSourcePin () : TOWSourcePin; virtual;

  procedure ReadStrData ( Reader : TReader );
  procedure WriteStrData( Writer : TWriter );

  procedure AddToAwaitsLinking();
  function  DoWrite() : Boolean;

  procedure DefineProperties(Filer: TFiler); override;

  procedure ClearPending();

public
  function  GetLinkStr( SaveFormat : Boolean ) : String;

public
  function  Notify( Operation : Integer; Data : Pointer ) : TOWNotificationResults; override;

public
  function  Connect( OtherPin : TOWPin ) : Boolean; override;
  procedure Disconnect(); override;
  
  function  IsCompatible ( const OtherPin : TOWPin ) : Boolean; override;

  function  CanConnectTo ( const OtherPin : TOWPin ) : Boolean; override;
  function  DependsOn ( const OtherPin : TOWPin ) : Boolean; override;
  function  IsConnected () : Boolean; override;
//  virtual TPinType GetAcceptedStreamTypes () const = 0;
//  virtual TPinType GetChainAcceptedStreamTypes () const;

  procedure Assign(Source: TPersistent); override;

public
  constructor Create(AOwner: TComponent);
  destructor  Destroy; override;

public
  function DoTryLink( Root : TComponent; SaveForm : Boolean ): Boolean;

public
  property SourcePin : TOWSourcePin read GetSourcePin write SetSourcePin;
  property DownStreamLinkName   : String read FDownStreamLinkName;
  property UpStreamLinkName     : String read FUpStreamLinkName;
  property IgnoreUpstream       : Boolean read FIgnoreUpstream write FIgnoreUpstream;

protected
  function  IsCompatibleChain ( const OtherPin : TOWPin ) : Boolean; override;
  
end;
//---------------------------------------------------------------------------
TOWPinList = class(TPersistent)
protected
  FPinsList             : TStringList;
  FLinkStorageSink      : TStringList;
  FLinkStorageSource    : TList;

public
  LastIndicatedCount : Integer;

public
  constructor Create( IsPinsOwner : Boolean );
  destructor Destroy; override;

public
  procedure AfterConstruction(); override;
  procedure BeforeDestruction(); override;
  
public
  function DoTryLink(): Boolean;

protected
  function  GetCount() : Integer;
  procedure SetCount( ACount : Integer ); virtual;
  function  GetPin( Item : Integer ) : TOWPin;
  function  GetName( Item : Integer ) : String;
  procedure SetName( Item : Integer; Name : String );

  procedure ClearPending();


protected
  procedure ReadStrData ( Reader : TReader ); virtual;
  procedure WriteStrData( Writer : TWriter );

//  procedure AddToAwaitsLinking();
  function  DoWrite() : Boolean; virtual;

  procedure DefineProperties(Filer: TFiler); override;

  function  GetLinkStrInt( Offset : String ) : String;

public
  procedure Add( Item : TOWPin );
  procedure AddNamed( Item : TOWPin; Name : String );
  procedure Remove( Item : TOWPin );
  procedure Delete( Index : Integer );

  procedure Clear();

public
  property Pins[ Item : Integer ] : TOWPin read GetPin;
  property Names[ Item : Integer ] : String read GetName write SetName;

published
  property Count : Integer read GetCount write SetCount;

public
  FPinsOwner : Boolean;

end;
//---------------------------------------------------------------------------
type TOWPinCreateFunction = function ( APinListOwner : TOWPinList ) : TOWPin of object;
type TOWPinDestroyFunction = procedure ( APinListOwner : TOWPinList ) of object;
//---------------------------------------------------------------------------
TOWPinListOwner = class(TOWPinList)
protected
  FPinCreateFunction  : TOWPinCreateFunction;
  FPinDestroyFunction : TOWPinDestroyFunction;

  FMinPins : Integer;
  FMaxPins : Integer;

protected
  procedure SetCount( ACount : Integer ); override;
  procedure ReadStrData ( Reader : TReader ); override;
  function  DoWrite() : Boolean; override;

public
  constructor Create( APinCreateFunction : TOWPinCreateFunction; APinDestroyFunction : TOWPinDestroyFunction = NIL );
  constructor CreateEx( AMinPins : Integer; AMaxPins : Integer; APinCreateFunction : TOWPinCreateFunction; APinDestroyFunction : TOWPinDestroyFunction = NIL );
//  destructor  Destroy; override;

end;
//---------------------------------------------------------------------------
type PPinEntry = ^TPinEntry;
//---------------------------------------------------------------------------

function  OWValueToString( StreamPin : TOWPin; Linking : String; AddRoot : Boolean; SaveValue : Boolean ) : String;
procedure OWGetPinsValueList( List : TStrings; StreamPin : TOWPin; Link : String; FullList : Boolean; SaveValue : Boolean );
procedure OWGetPinsValueListSingle( List : TStrings; OwnerComponent : TComponent;  StreamPin : TOWPin; Link : String; FullList : Boolean; RootName : String; SaveValue : Boolean );
procedure OWGetPinsValueListSingleRoot( List : TStrings; OwnerComponent : TComponent;  StreamPin : TOWPin; Link : String; FullList : Boolean; RootName : String; SaveValue : Boolean );
function  OWGetMainOwnerComponent( Component : TComponent ) : TComponent;
procedure OWRegisterWireStream( ID : TGUID; Name : String );
function  OWGetStreamTypeFromID( ID : TGUID ) : String;
procedure OWLinkAwaitsLinking( Root : TComponent );
procedure OWReportAwaitsLinking();
function  OWGetAllLinked() : Boolean;
procedure OWGetClassPropertiesOfTypeName ( AClass : TObject; TypeName : String; AStrings : TStrings; SaveValue : Boolean );
procedure OWClearPendingLinks();

type TOWPinNotifier = class
  public
    procedure AddPin( APin : TOWPin ); virtual; abstract;
    procedure RemovePin( APin : TOWPin ); virtual; abstract;
    procedure ChangePin( APin : TOWPin ); virtual; abstract;

    procedure AddPinList( APinList : TOWPinList ); virtual; abstract;
    procedure RemovePinList( APinList : TOWPinList ); virtual; abstract;
    procedure ChangePinList( APinList : TOWPinList ); virtual; abstract;

end;

procedure OWAddPinNotifier( ANotifyItem : TOWPinNotifier );
procedure OWRemovePinNotifier( ANotifyItem : TOWPinNotifier );

procedure Register;

implementation

  uses Dialogs, Forms;

type
TStreamTypeEntry=record
  ID    : TGUID;
  Name  : String;
end;

var
  AwaitsLinkingSourcePins       : TList;
  AwaitsLinkingSinkPins         : TList;
  AwaitsLinkingLists            : TList;

  StreamTypes                   : array of TStreamTypeEntry;
  IsInLinking                   : Boolean;
  AllLinked                     : Boolean;

  NotifyList                    : TList;

//---------------------------------------------------------------------------
procedure OWAddPinNotifier( ANotifyItem : TOWPinNotifier );
begin
  NotifyList.Add( ANotifyItem ); 
end;
//---------------------------------------------------------------------------
procedure OWRemovePinNotifier( ANotifyItem : TOWPinNotifier );
begin
  NotifyList.Remove( ANotifyItem );
end;
//---------------------------------------------------------------------------
procedure OWNotifyAddPin( APin : TOWPin );
var
  I : Integer;

begin
  for I := 0 to NotifyList.Count - 1 do
    TOWPinNotifier( NotifyList.Items[ I ] ).AddPin( APin );
    
end;
//---------------------------------------------------------------------------
procedure OWNotifyRemovePin( APin : TOWPin );
var
  I : Integer;

begin
  for I := 0 to NotifyList.Count - 1 do
    TOWPinNotifier( NotifyList.Items[ I ] ).RemovePin( APin );

end;
//---------------------------------------------------------------------------
procedure OWNotifyChangePin( APin : TOWPin );
var
  I : Integer;

begin
  for I := 0 to NotifyList.Count - 1 do
    TOWPinNotifier( NotifyList.Items[ I ] ).ChangePin( APin );
    
end;
//---------------------------------------------------------------------------
procedure OWNotifyAddPinList( APinList : TOWPinList );
var
  I : Integer;

begin
  for I := 0 to NotifyList.Count - 1 do
    TOWPinNotifier( NotifyList.Items[ I ] ).AddPinList( APinList );
    
end;
//---------------------------------------------------------------------------
procedure OWNotifyRemovePinList( APinList : TOWPinList );
var
  I : Integer;

begin
  for I := 0 to NotifyList.Count - 1 do
    TOWPinNotifier( NotifyList.Items[ I ] ).RemovePinList( APinList );

end;
//---------------------------------------------------------------------------
procedure OWNotifyChangePinList( APinList : TOWPinList );
var
  I : Integer;

begin
  for I := 0 to NotifyList.Count - 1 do
    TOWPinNotifier( NotifyList.Items[ I ] ).ChangePinList( APinList );
    
end;
//---------------------------------------------------------------------------
procedure OWRegisterWireStream( ID : TGUID; Name : String );
var
  Entry : TStreamTypeEntry;

begin
  Entry.ID := ID;
  Entry.Name := Name;
  SetLength( StreamTypes, Length( StreamTypes ) + 1 );
  StreamTypes[ Length( StreamTypes ) - 1 ] := Entry;
end;
//---------------------------------------------------------------------------
function EqualGUIDs( ID1 : TGUID; ID2 : TGUID ) : Boolean;
var
  I : Integer;

begin
  Result := False;
  if( ID1.D1 <> ID2.D1 ) then
    Exit;

  if( ID1.D2 <> ID2.D2 ) then
    Exit;

  if( ID1.D3 <> ID2.D3 ) then
    Exit;

  for I := 0 to 7 do
    if( ID1.D4[ I ] <> ID2.D4[ I ] ) then
      Exit;

  Result := True;
end;
//---------------------------------------------------------------------------
function  OWGetStreamTypeFromID( ID : TGUID ) : String;
var
  I : Integer;
  
begin
  for I := 0 to Length( StreamTypes ) - 1 do
    if( EqualGUIDs ( StreamTypes[ i ].ID, ID ))then
      begin
      Result := StreamTypes[ i ].Name;
      Exit;
      end;

  Result := '';

end;
//---------------------------------------------------------------------------
function OWGetMainOwnerComponent ( Component : TComponent ) : TComponent;
begin
  if ( Component.Owner = NIL ) then
    Result := Component

  else if ( Component.Owner is TApplication ) then
    Result := Component
    
  else if( Component is TDataModule ) then
    Result := Component

  else if( Component is TCustomForm ) then
    Result := Component

  else
    Result := OWGetMainOwnerComponent ( Component.Owner );
    
end;

//---------------------------------------------------------------------------

procedure RemoveFromAwaitsLinking( Pin : TOWPin );
begin
  AwaitsLinkingSourcePins.Remove( Pin );
  AwaitsLinkingSinkPins.Remove( Pin );
  AwaitsLinkingLists.Remove( Pin );

end;

//---------------------------------------------------------------------------
function  OWGetAllLinked() : Boolean;
begin
  Result := AllLinked;
end;
//---------------------------------------------------------------------------
procedure OWReportAwaitsLinking();
var
  Info  : String;
  I     : Integer;
  SinkPin       : TOWSinkPin;
  SourcePin     : TOWSourcePin;
  PinList       : TOWPinList;

begin
  Info := 'AwaitsLinkingSourcePins =' + IntToStr( AwaitsLinkingSourcePins.Count ) + #13;
  for I := 0 to AwaitsLinkingSourcePins.Count - 1 do
    begin
    SourcePin := TOWSourcePin( AwaitsLinkingSourcePins.Items[ I ] );
    Info := Info + '  ' + SourcePin.GetLinkStrInt( '    ' ) + #13;
    end;

  Info := Info + 'AwaitsLinkingSinkPins =' + IntToStr( AwaitsLinkingSinkPins.Count ) + #13;
  for I := 0 to AwaitsLinkingSinkPins.Count - 1 do
    begin
    SinkPin := TOWSinkPin( AwaitsLinkingSinkPins.Items[ I ] );
    Info := Info + '  ' + SinkPin.GetLinkStr( True ) + #13;
    end;

  Info := Info + 'AwaitsLinkingLists =' + IntToStr( AwaitsLinkingLists.Count );
  for I := 0 to AwaitsLinkingLists.Count - 1 do
    begin
    PinList := TOWPinList( AwaitsLinkingLists.Items[ I ] );
    Info := Info + '  ' + PinList.GetLinkStrInt( '    ' ) + #13;
//    Info := Info + '  ' + 'X' + #13;
    end;
     
          
  ShowMessage( Info );
end;
//---------------------------------------------------------------------------
procedure OWClearPendingLinks();
var
  I             : Integer;
  SinkPin       : TOWSinkPin;
  SourcePin     : TOWSourcePin;
  PinList       : TOWPinList;

begin
  I := 0;
  while ( I < AwaitsLinkingSinkPins.Count ) do
    begin
    SinkPin := TOWSinkPin( AwaitsLinkingSinkPins.Items[ I ] );
    SinkPin.ClearPending();
    Inc( I );
    end;

  I := 0;
  while ( I < AwaitsLinkingSourcePins.Count ) do
    begin
    SourcePin := TOWSourcePin( AwaitsLinkingSourcePins.Items[ I ] );
    SourcePin.ClearPending();
    Inc( I );
    end;

  I := 0;
  while ( I < AwaitsLinkingLists.Count ) do
    begin
    PinList := TOWPinList( AwaitsLinkingLists.Items[ I ] );
    PinList.ClearPending();
    Inc( I );
    end;

  AwaitsLinkingLists.Clear();
  AwaitsLinkingSinkPins.Clear();
  AwaitsLinkingSourcePins.Clear();
  AllLinked := True;

end;
//---------------------------------------------------------------------------
procedure LinkAwaitsLinkingSingleForm( Root : TComponent );
var
  I             : Integer;
  SinkPin       : TOWSinkPin;
  SourcePin     : TOWSourcePin;
  PinList       : TOWPinList;
  
begin
  if( Root <> NIL ) then
    begin
    Root := OWGetMainOwnerComponent ( Root );
    I := 0;
    while ( I < AwaitsLinkingSinkPins.Count ) do
      begin
      SinkPin := TOWSinkPin( AwaitsLinkingSinkPins.Items[ I ] );
      if SinkPin.DoTryLink( Root, True ) then
        begin
        AwaitsLinkingSinkPins.Delete( I );
        Continue;
        end;

      Inc( I );
      end;

    I := 0;
    while ( I < AwaitsLinkingSourcePins.Count ) do
      begin
      SourcePin := TOWSourcePin( AwaitsLinkingSourcePins.Items[ I ] );
      if SourcePin.DoTryLink( Root, True ) then
        begin
        AwaitsLinkingSourcePins.Delete( I );
        Continue;
        end;

      Inc( I );
      end;
    end;

  I := 0;
  while ( I < AwaitsLinkingLists.Count ) do
    begin
    PinList := TOWPinList( AwaitsLinkingLists.Items[ I ] );
    if PinList.DoTryLink() then
      begin
      AwaitsLinkingLists.Delete( I );
      Continue;
      end;

    Inc( I );
    end;

  if( ( AwaitsLinkingLists.Count = 0 ) and
      ( AwaitsLinkingSourcePins.Count = 0 ) and
      ( AwaitsLinkingSinkPins.Count = 0 )) then
      
    AllLinked := True;

end;

//---------------------------------------------------------------------------
procedure OWLinkAwaitsLinking( Root : TComponent );
var
  MainRoot      : TComponent;
  I             : Integer;
//  Loaded        : Boolean;
  CurrentRoot   : TComponent;

begin
  if( AllLinked ) then
    Exit;
    
  if( IsInLinking ) then
    Exit;

  IsInLinking := True;

  if( Root = NIL ) then
    begin
    LinkAwaitsLinkingSingleForm( NIL );
    IsInLinking := False;
    Exit;
    end;

  if( csDesigning in Root.ComponentState ) then
    begin
    MainRoot := OWGetMainOwnerComponent ( Root );
    if( MainRoot.Owner <> NIL ) then
      if( MainRoot.Owner is TApplication ) then
        begin
        MainRoot := MainRoot.Owner;
//        Loaded := True;
        for I := 0 to MainRoot.ComponentCount - 1 do
          begin
          CurrentRoot := MainRoot.Components[ I ];
          if(( CurrentRoot is TCustomForm ) or
             ( CurrentRoot is TDataModule )) then
            begin
//            if( ( CurrentRoot.ComponentState * [ csLoading, csReading, csUpdating, csFixups ] ) <> [] ) then
//              Loaded := False;

            LinkAwaitsLinkingSingleForm( CurrentRoot );
            end;

          end;

//        if( Loaded ) then
//          ClearPendingLinks();

        IsInLinking := False;
        Exit;
        end;

    LinkAwaitsLinkingSingleForm( Root );
    IsInLinking := False;
    Exit;
    end;

//  Loaded := True;
  for I := 0 to Screen.CustomFormCount - 1 do
    begin
    CurrentRoot := Screen.CustomForms[ I ];
//    if( ( CurrentRoot.ComponentState * [ csLoading, csReading, csUpdating, csFixups ] ) <> [] ) then
//      Loaded := False;

    LinkAwaitsLinkingSingleForm( CurrentRoot );
    end;

  for I := 0 to Screen.DataModuleCount - 1 do
    begin
    CurrentRoot := Screen.DataModules[ I ];
//    if( ( CurrentRoot.ComponentState * [ csLoading, csReading, csUpdating, csFixups ] ) <> [] ) then
//      Loaded := False;

    LinkAwaitsLinkingSingleForm( CurrentRoot );
    end;

////  if( Loaded ) then
////    ClearPendingLinks();

  IsInLinking := False;
  Exit;

{
  MainRoot := GetMainOwnerComponent ( Root );
  if( MainRoot.Owner <> NIL ) then
    if( MainRoot.Owner is TApplication ) then
      begin
      MainRoot := MainRoot.Owner;
      Loaded := True;
      for I := 0 to MainRoot.ComponentCount - 1 do
          begin
          CurrentRoot := MainRoot.Components[ I ]; 
          if(( CurrentRoot is TCustomForm ) or
             ( CurrentRoot is TDataModule )) then
            begin
            if( ( CurrentRoot.ComponentState * [ csLoading, csReading, csUpdating, csFixups ] ) = [] ) then
              Loaded := False;
            
            LinkAwaitsLinkingSingleForm( MainRoot.Components[ I ] );
            end;
            
          end;

      if( Loaded ) then
        ClearPendingLinks(); 

      IsInLinking := False;
      Exit;
      end;

}
end;
//---------------------------------------------------------------------------
constructor TOWPin.Create(AOwner: TComponent);
begin
  inherited Create;
  StreamType := TPinType.Create;
  FOwner := AOwner;
  FDeleteNotifyes := TList.Create;
  PinsNeedRefresh := True;
  InSending := False;
end;
//---------------------------------------------------------------------------
destructor TOWPin.Destroy;
var
  I : Integer;

begin
  RemoveFromAwaitsLinking( Self );
  for I := 0 to FDeleteNotifyes.Count - 1 do
    TOWPin( FDeleteNotifyes.Items[ I ] ).PinDeletedNotify( Self );

  FDeleteNotifyes.Free();
  StreamType.Free();
  inherited;
end;
//---------------------------------------------------------------------------
procedure TOWPin.AfterConstruction();
begin
  inherited;
  OWNotifyAddPin( Self );
end;
//---------------------------------------------------------------------------
procedure TOWPin.BeforeDestruction();
begin
  OWNotifyRemovePin( Self );
  inherited;
end;
//---------------------------------------------------------------------------
procedure TOWPin.AddDeleteNotify( ADeleteNotifyPin : TOWPin );
begin
  if( FDeleteNotifyes.IndexOf( ADeleteNotifyPin ) = -1 ) then
    begin
    FDeleteNotifyes.Add( ADeleteNotifyPin );
    ADeleteNotifyPin.AddDeleteNotify( Self );
    end;

end;
//---------------------------------------------------------------------------
{
procedure TOWPin.RemoveDeleteNotify( ADeleteNotifyPin : TOWPin );
begin
  FDeleteNotifyes.Remove( ADeleteNotifyPin );
end;
}
//---------------------------------------------------------------------------
procedure TOWPin.PinDeletedNotify( ADeletedPin : TOWPin );
begin
  FDeleteNotifyes.Remove( ADeletedPin ); 
end;
//---------------------------------------------------------------------------
procedure TOWPin.AddType( ID : TGUID );
begin
  StreamType.AddType( ID, DefaultSubmit );
end;

//---------------------------------------------------------------------------
procedure TOWPin.AddType( ID : TGUID; SubmitFunction : TOWSubmitFunction );
begin
  StreamType.AddType( ID, SubmitFunction );
end;

//---------------------------------------------------------------------------

function  TOWPin.DefaultSubmit( Handler : IOWStream; Operation : Integer; Data : Pointer; State : TOWNotificationState ) : TOWNotificationResults;
begin
  Result := [];
end;

//---------------------------------------------------------------------------

function  TOWPin.GetSubmitFunction ( const OtherPin : TOWPin ) : TOWSubmitFunction;
var
  Unknown : IUnknown;
  I       : Integer;

begin
  Result := NIL;
  if( OtherPin <> NIL ) then
    begin
    for I := 0 to StreamType.Count - 1 do
      begin
      if ( OtherPin.QueryInterface( StreamType.ID[ I ], Unknown ) = S_OK ) then
        begin
        Result := StreamType.SubmitFunc[ I ];
        Exit;
        end;
      end;
    end;

end;
//---------------------------------------------------------------------------
function  TOWPin.GetLinkName ( const OtherPin : TOWPin ) : String;
var
  Unknown : IUnknown;
  I       : Integer;

begin
  if( StreamType.Count = 0 ) then
    begin
    Result := '';
    Exit;
    end;

  Result := 'Unregistered stream';
  if( OtherPin <> NIL ) then
    begin
    for I := 0 to StreamType.Count - 1 do
      begin
      if ( OtherPin.QueryInterface( StreamType.ID[ I ], Unknown ) = S_OK ) then
        begin
        Result := OWGetStreamTypeFromID( StreamType.ID[ I ] );
        Exit;
        end;
      end;
    end;

end;
//---------------------------------------------------------------------------

function  TOWPin.TypeCheckInterfaces ( const OtherPin : TOWPin ) : Boolean;
var
  Unknown : IUnknown;
  I       : Integer;

begin
  Result := True;
  
  if( StreamType.Count = 0 ) then
    Exit;

  Result := False;
  if( OtherPin <> NIL ) then
    begin
    for I := 0 to StreamType.Count - 1 do
      begin
      if ( OtherPin.QueryInterface( StreamType.ID[ I ], Unknown ) = S_OK ) then
        begin
        Result := True;
        Exit;
        end;
      end;
    end;

end;
//---------------------------------------------------------------------------
function  TOWPin.CheckInterfaces ( const OtherPin : TOWPin ) : Boolean;
begin
  Result:= False;
  
  if( OtherPin <> NIL ) then
    if ( TypeCheckInterfaces( OtherPin )) then
      Result:= OtherPin.TypeCheckInterfaces( Self )

end;
//---------------------------------------------------------------------------

function  TOWPin.IsCompatibleChain ( const OtherPin : TOWPin ) : Boolean;
begin
  Result := IsCompatible ( OtherPin );
end;
//---------------------------------------------------------------------------
function  TOWPin.CanConnectTo ( const OtherPin : TOWPin ) : Boolean;
begin
  Result := IsCompatibleChain( OtherPin );
end;
//---------------------------------------------------------------------------
function TOWPin.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface(IID, Obj) then
    Result := S_OK

  else
    Result := E_NOINTERFACE
end;

function TOWPin._AddRef: Integer; stdcall;
begin
  _AddRef := 1;
end;

function TOWPin._Release: Integer; stdcall;
begin
  _Release := 1;
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
constructor TOWSourcePin.Create(AOwner: TComponent);
begin
  inherited;
  InDependOn := False;
  SinkPins   := TList.Create;
  FFunctionSources := TList.Create;
  FDataTypeSources := TList.Create;
  FLinkStorage := NIL;

end;

//---------------------------------------------------------------------------
destructor TOWSourcePin.Destroy;
begin
  OWLinkAwaitsLinking( Owner );
  Disconnect();

  SinkPins.Free;
  FDataTypeSources.Free;
  FFunctionSources.Free;

  if( CurrentEditorPtr <> NIL )then
    CurrentEditorPtr^ := NIL;

  if( FLinkStorage <> NIL ) then
    FLinkStorage.Free();

  inherited;
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.ReadStrData( Reader : TReader );
var
  Ident   : String;

begin
  Reader.ReadListBegin();
  while not Reader.EndOfList do
    begin
    Ident := Reader.ReadIdent();
    TryLinkTo ( Reader.Root, Ident, True, True );
//        AddToStorage( Ident );
{
    if( Ident <> 'nil' ) then
      begin
      FLinkStorage.Add( Ident );
      end

    else
      FLinkStorage.Add( '' );
}
    end;

  Reader.ReadListEnd();

end;
//---------------------------------------------------------------------------
function  TOWSourcePin.DoTryLink( Root : TComponent; SaveForm : Boolean ): Boolean;
var
  I : Integer;
  
begin
  Result := True;
  if( FLinkStorage = NIL ) then
    Exit;
  
  I := 0;
  while ( I < FLinkStorage.Count ) do
    if( TryLinkTo( Root, FLinkStorage.Strings[ I ], False, True )) then
      begin
//      ShowMessage( FLinkStorage.Strings[ I ] + ' Linked !' ); 
      FLinkStorage.Delete( I );
      end

    else
      begin
      Inc( I );
      end;

  if ( FLinkStorage.Count = 0 ) then
    begin
    FLinkStorage.Free();
    FLinkStorage := NIL;
    Result := True;
    Exit;
    end;

  Result := False;

end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.TryLinkToList ( Root : TComponent; SinkNames : TStringList );
var
  I : Integer;
  
begin
  for I := 0 to SinkNames.Count - 1 do
    TryLinkTo ( Root, SinkNames.Strings[ I ], True, True );
  
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.ClearPending();
begin
  if( FLinkStorage <> NIL ) then
    FLinkStorage.Free();

  FLinkStorage := NIL;
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.TryLinkTo ( Root : TComponent; SinkName : String; TryLater : Boolean; SaveForm : Boolean ) : Boolean;
var
  Values        : TStringList;
  SinkPin       : TOWSinkPin;
  I             : Integer;
  RootName      : String;
  MyRoot        : TComponent;

begin
  Result := True;
  if( SinkName = '' ) then
    begin
    if( not SaveForm ) then
      Disconnect();
//      Connect( NIL );

    Exit;
    end;

  Result := False;

  Values := TStringList.Create();
  try
//    ShowMessage( Designer.GetRoot().Name );
    MyRoot := OWGetMainOwnerComponent( Owner );
    if( ( MyRoot <> Root ) and ( Root <> NIL )) then
      begin
      RootName := Root.Name;
      OWGetPinsValueListSingle( Values, Root, Self, '.', True, RootName, SaveForm );
      end

    else 
      OWGetPinsValueList( Values, Self, '.', True, SaveForm );
      
//    GetPinsValueListSingle( Values, FOwner, Self, '.', True, '', SaveForm );
//    ShowMessage( SinkName );
//    ShowMessage( Values.Text );
    for I := 0 to Values.Count - 1 do
      begin
      if( Values.Strings[ I ] = SinkName ) then
        begin
        SinkPin := TOWSinkPin( Values.Objects[ i ] );
        Connect( SinkPin );
        Result := True;
        Break;
        end;
        
      end;

  finally
    Values.Free();
    if( not Result ) then
      if( TryLater ) then
        AddToStorage( SinkName );

  end;
{
  if ( not DoTryLink( Root, SaveForm )) then
    begin
    if( TryLater ) then
      AddToAwaitsLinking() // Add to be updated later;

    else
      FLinkStorage := '';

    Result := False;
    Exit;
    end;
}
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.AddToStorage( Ident : String );
begin
//  ShowMessage( Ident );
  if( FLinkStorage = NIL ) then
    FLinkStorage := TStringList.Create;

  FLinkStorage.Sorted := True;
  FLinkStorage.Duplicates := dupIgnore;
  FLinkStorage.Add( Ident );
  if( AwaitsLinkingSourcePins.IndexOf( Self ) = -1 ) then
    begin
    AwaitsLinkingSourcePins.Add( Self );
    AllLinked := False;
    end;
    
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.WriteStrData( Writer : TWriter );
var
  I      : Integer;

begin
  OWLinkAwaitsLinking( Owner );
  Writer.WriteListBegin();
  for I := 0 to SinkCount - 1 do
    begin
    Writer.WriteIdent ( GetLinkStr( I, True ));
{
    if( Sinks[ I ] is TOWSinkPin ) then
      if( TOWSinkPin( Pins[ I ] ).DoWrite() ) then
        begin
        TOWSinkPin( Pins[ I ] ).WriteStrData( Writer );
        Continue;
        end;
        
    Writer.WriteIdent( 'nil' );
}
    end;
    
  Writer.WriteListEnd();

end;
//---------------------------------------------------------------------------
function  TOWSourcePin.DoWrite() : Boolean;
begin
  Result := ( IsConnected() ) or ( FLinkStorage <> NIL );
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty( 'SinkPins', ReadStrData, WriteStrData, DoWrite());
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.GetLinkStr( Item : Integer; SaveFormat : Boolean  ) : String;
var
  AddRoot : Boolean;

begin
  AddRoot := False;
  if( SinkCount > 0 ) then
    AddRoot := OWGetMainOwnerComponent( Sinks[ Item ].Owner ) <> OWGetMainOwnerComponent( Owner );
     
  Result := OWValueToString( Sinks[ Item ], '.', AddRoot, SaveFormat );
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.GetLinkStrInt( Offset : String ) : String;
var
  I : Integer;
  
begin
  Result := '';

  if( FLinkStorage <> NIL ) then
    begin
    for I := 0 to FLinkStorage.Count - 1 do
      Result := Result + Offset + FLinkStorage.Strings[ I ] + #13;
      
    end;

end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.IntConnect( SinkPin : TOWSinkPin );
var
  PEntry : PPinEntry;
  
begin
  New( PEntry );
  PEntry.SinkPin := SinkPin;
  PEntry.SubmitFunction := GetSubmitFunction ( SinkPin );
  PEntry.ModificationLevel := 0;
  SinkPins.Add( PEntry );
end;

//---------------------------------------------------------------------------
procedure TOWSourcePin.IntDisconnect( SinkPin : TOWSinkPin );
var
  Index : Integer;
  PEntry : PPinEntry;

begin
  for Index := 0 to GetSinkCount() - 1 do
    begin
    PEntry := PPinEntry( SinkPins.Items[ Index ]); 
    if( PEntry.SinkPin = SinkPin ) then
      begin
      Dispose( PEntry );      
      SinkPins.Delete( Index );
      OWNotifyChangePin( Self );
      Exit;
      end;
    end;
     
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.Notify( Operation : Integer; Data : Pointer ) : TOWNotificationResults;
var
  I      : Integer;
  State  : TOWNotificationState;
  Status : TOWNotificationResults;
  
begin
  if( InSending ) then
    begin
    raise EOWException.Create( 'The pin depends on it self. Please protect with functional dependancy.' );
    Exit;
    end;
    
  InSending := True;
  Result := [];

  for I := 0 to GetSinkCount() - 1 do
    begin
    if( I < GetSinkCount() - 1 ) then
      State := []

    else
      State := [nsLastIteration];

    Status := TOWSubmitFunction( GetSubmit( I ))( GetSink ( I ), Operation, Data, State );

    if( nrDataChanged in Status ) then
      SinkChangedData( I );

    end;

  ReorderChangedData();

  InSending := False;
end;
//---------------------------------------------------------------------------
function TOWSourcePin.GetSinkCount() : Integer; // const;
begin
  OWLinkAwaitsLinking( Owner );
  Result := SinkPins.Count;
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.GetSinkCountAll() : Integer; // const;
begin
  Result := GetSinkCount();
  if( FLinkStorage <> NIL ) then
    Inc( Result, FLinkStorage.Count );
     
end;
//---------------------------------------------------------------------------
function TOWSourcePin.GetSink( Index : Integer ) : TOWSinkPin;
begin
  OWLinkAwaitsLinking( Owner );
  Result := PPinEntry( SinkPins.Items[ Index ]).SinkPin;
end;
//---------------------------------------------------------------------------
function TOWSourcePin.GetSubmit( Index : Integer ) : TOWSubmitFunction;
begin
  OWLinkAwaitsLinking( Owner );
  Result := PPinEntry( SinkPins.Items[ Index ]).SubmitFunction;
end;
//---------------------------------------------------------------------------
function PinSortMethod(Item1, Item2: Pointer): Integer;
begin
  Result := PPinEntry( Item1 ).ModificationLevel - PPinEntry( Item2 ).ModificationLevel;
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.ReorderChangedData();
var
  PEntry : PPinEntry;
  I      : Integer;

begin
  for I:= 0 to SinkPins.Count - 1 do
    begin
    PEntry := PPinEntry( SinkPins.Items[ I ]);
    if( PEntry.ModificationLevel > 0 ) then
      Dec( PEntry.ModificationLevel );
      
    end;
    
  SinkPins.Sort( PinSortMethod );
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.SinkChangedData( Index : Integer );
var
  PEntry : PPinEntry;
  
begin
  OWLinkAwaitsLinking( Owner );
  PEntry := PPinEntry( SinkPins.Items[ Index ]);
  if( PEntry.ModificationLevel < 10000 ) then
    Inc( PEntry.ModificationLevel, 4 );

end;
//---------------------------------------------------------------------------

function TOWSourcePin.GetFunctionSourceCount() : Integer; // const;
begin
  Result := FFunctionSources.Count;
end;

//---------------------------------------------------------------------------

function TOWSourcePin.GetFunctionSources( Index : Integer ) : TOWSinkPin;
begin
  Result := FFunctionSources.Items[ Index ];  
end;

//---------------------------------------------------------------------------

function TOWSourcePin.GetDataTypeSourceCount() : Integer; // const;
begin
  Result := FDataTypeSources.Count;
end;

//---------------------------------------------------------------------------

function TOWSourcePin.GetDataTypeSources( Index : Integer ) : TOWSinkPin;
begin
  Result := FDataTypeSources.Items[ Index ];  
end;

//---------------------------------------------------------------------------
procedure TOWSourcePin.PinDeletedNotify( ADeletedPin : TOWPin );
begin
  inherited;
  FFunctionSources.Remove( ADeletedPin );
  FDataTypeSources.Remove( ADeletedPin );
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.IntAddFunctionSource( FunctionSourcePin : TOWSinkPin );
begin
  FFunctionSources.Add( FunctionSourcePin );
  FunctionSourcePin.AddDeleteNotify( Self );    // Tell me if you are gone so te remove you from the list. 
end;

//---------------------------------------------------------------------------

procedure TOWSourcePin.IntRemoveFunctionSource( FunctionSourcePin : TOWSinkPin );
begin
  FFunctionSources.Remove( FunctionSourcePin );
end;

//---------------------------------------------------------------------------

procedure TOWSourcePin.IntAddDataTypeSource( DataTypeSourcePin : TOWSinkPin );
begin
  FDataTypeSources.Add( DataTypeSourcePin );
  DataTypeSourcePin.AddDeleteNotify( Self );    // Tell me if you are gone so te remove you from the list. 
end;

//---------------------------------------------------------------------------

procedure TOWSourcePin.IntRemoveDataTypeSource( DataTypeSourcePin : TOWSinkPin );
begin
  FDataTypeSources.Remove( DataTypeSourcePin );
end;
//---------------------------------------------------------------------------

function TOWSourcePin.Connect( OtherPin : TOWPin ) : Boolean;
begin
  Result := False;
  if( OtherPin <> NIL ) then
    if( OtherPin is TOWSinkPin ) then
      Result := TOWSinkPin( OtherPin ).Connect( Self );

//  if( Result ) then
//    NewConnection( SinkPin, GetSubmitFunction( SinkPin ));  

end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.NewConnection( ASinkPin : TOWSinkPin; ASubmitFunction : TOWSubmitFunction );
//var SinkRoot : TComponent;
begin
  if( InSending ) then
    begin
    raise EOWException.Create( 'The pin depends on it self. Please protect with functional dependancy.' );
    Exit;
    end;
    
  InSending := True;
  ASubmitFunction( ASinkPin, 0, NIL, [ nsNewLink, nsLastIteration ] );
  OWNotifyChangePin( Self );
  InSending := False;
{
  SinkRoot := GetMainOwnerComponent( ASinkPin.Owner );
  if( SinkRoot is TCustomForm ) then
    if( Assigned( TCustomForm( SinkRoot ).Designer )) then
      TCustomForm( SinkRoot ).Designer.Modified;
}
end;
//---------------------------------------------------------------------------
function TOWSourcePin.ChainCheckInterfaces( ASinkPin : TOWSinkPin ) : Boolean;
//var
//  I     : Integer;
   
begin
  Result := True;       // Not available in this Delphi version. Will be implemented in the future.
  Exit;

{
  for ( TAceptorsList::const_iterator Iter = FDataTypeSources.begin(); Iter != FDataTypeSources.end(); Iter ++ )
    ConcatStreamType.ConcatTypes ( ( *Iter )->GetChainAcceptedStreamTypes ());
}
//  for I := 0 to FDataTypeSources.Count -1 do 
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.IsCompatible ( const OtherPin : TOWPin ) : Boolean;
var
  SinkPin : TOWSinkPin;
  
begin
  Result := False;
  if( OtherPin is TOWSinkPin ) then
    begin
    SinkPin := TOWSinkPin( OtherPin );
    Result := SinkPin.IsCompatible ( Self );
    end;

end;
//---------------------------------------------------------------------------
function  TOWSourcePin.CanConnectTo ( const OtherPin : TOWPin ) : Boolean;
begin
  Result := False;
  if ( OtherPin <> NIL ) then
    begin
      if ( OtherPin <> Self ) then
        begin
          if ( OtherPin is TOWSinkPin ) then
            Result := OtherPin.CanConnectTo( Self );
        end;
      end;

end;
//---------------------------------------------------------------------------
function  TOWSourcePin.IsConnected () : Boolean;
begin
  Result := ( GetSinkCount() > 0 ); 
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.DependsOn ( const OtherPin : TOWPin ) : Boolean;
var
  I : Integer;
  
begin
  Result := False;
  if( InDependOn ) then
    Exit;

  InDependOn := True;
  if ( OtherPin <> NIL ) then
    begin

    Result := True;
    for I := 0 to FFunctionSources.Count - 1 do
      if( FFunctionSources.Items[ I ] = OtherPin ) then
        begin
        Result := True;
        InDependOn := False;
        Exit;
        end;

    for I := 0 to FFunctionSources.Count - 1 do
      if( TOWPin( FFunctionSources.Items[ I ] ).DependsOn ( OtherPin ) ) then
        begin
        InDependOn := False;
        Exit;
        end;

    end;

  Result := False;
  InDependOn := False;
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.IsLinkedTo ( OtherPin : TOWPin ) : Boolean; //const;
var
  I     : Integer;
  
begin
  Result := True;
  for I := 0 to GetSinkCount() - 1 do
    if( OtherPin = GetSink( I ) ) then
      Exit;
      
  Result := False;
end;
//---------------------------------------------------------------------------
function  TOWSourcePin.IsLinkedTo( SinkName : String ) : Boolean;
var
  I     : Integer;
  
begin
  Result := True;
  for I := 0 to GetSinkCount() - 1 do
    if( SinkName = GetLinkStr( I, False )) then
      Exit;

  if( FLinkStorage <> NIL ) then
    begin
    for I := 0 to FLinkStorage.Count - 1 do 
      if( FLinkStorage.Strings[ I ] = SinkName ) then
        Exit;
    end;
     
  Result := False;
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.Disconnect ();
var
  I         : Integer;
  SinkPin   : TOWSinkPin;
  PEntry    : PPinEntry;

begin
  for I := 0 to GetSinkCount() - 1 do
    begin
    SinkPin := GetSink( I );
    SinkPin.FSourcePin := NIL;
//    GetSink( I ).FSourcePin := NIL;
    PEntry := PPinEntry( SinkPins.Items[ I ]);
    Dispose( PEntry );
    end;

  if( FLinkStorage <> NIL ) then
    begin
    FLinkStorage.Free();
    FLinkStorage := NIL;
    end;
    
  SinkPins.Clear;
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.AddFunctionSource ( FunctionSourcePin : TOWSinkPin );
begin
  FunctionSourcePin.IntAddFunctionDependance( Self );
  IntAddFunctionSource ( FunctionSourcePin );
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.RemoveFunctionSource ( FunctionSourcePin : TOWSinkPin );
begin
  IntRemoveFunctionSource ( FunctionSourcePin );
  FunctionSourcePin.IntRemoveFunctionDependance( Self );
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.AddDataTypeSource ( DataTypeSourcePin : TOWSinkPin );
begin
  DataTypeSourcePin.IntAddDataTypeDependance( Self );
  IntAddDataTypeSource ( DataTypeSourcePin );
end;
//---------------------------------------------------------------------------
procedure TOWSourcePin.RemoveDataTypeSource ( DataTypeSourcePin : TOWSinkPin );
begin
  IntRemoveDataTypeSource ( DataTypeSourcePin );
  DataTypeSourcePin.IntRemoveDataTypeDependance( Self );
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
constructor TOWSinkPin.Create(AOwner: TComponent);
begin
  inherited;
  FSourcePin := NIL;
  FFunctionDependants := TList.Create;
  FDataTypeDependants := TList.Create;
end;
//---------------------------------------------------------------------------
destructor TOWSinkPin.Destroy;
begin
  Connect( NIL );
  OWLinkAwaitsLinking( Owner );
  FFunctionDependants.Free;
  FDataTypeDependants.Free;
  
  if( CurrentEditorPtr <> NIL )then
    CurrentEditorPtr^ := NIL;
    
  inherited;
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.PinDeletedNotify( ADeletedPin : TOWPin );
begin
  inherited;
  FFunctionDependants.Remove( ADeletedPin );
  FDataTypeDependants.Remove( ADeletedPin );
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.IntAddFunctionDependance ( FunctionDependancePin : TOWSourcePin );
begin
  FFunctionDependants.Add( FunctionDependancePin );
  FunctionDependancePin.AddDeleteNotify( Self );    // Tell me if you are gone so te remove you from the list.
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.IntRemoveFunctionDependance ( FunctionDependancePin : TOWSourcePin );
begin
  FFunctionDependants.Remove( FunctionDependancePin );
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.IntAddDataTypeDependance ( DataTypeDependancePin : TOWSourcePin );
begin
  FDataTypeDependants.Add( DataTypeDependancePin );
  DataTypeDependancePin.AddDeleteNotify( Self );    // Tell me if you are gone so te remove you from the list.
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.IntRemoveDataTypeDependance ( DataTypeDependancePin : TOWSourcePin );
begin
  FDataTypeDependants.Remove( DataTypeDependancePin );
end;
//---------------------------------------------------------------------------
function TOWSinkPin.Notify( Operation : Integer; Data : Pointer ) : TOWNotificationResults;
begin
  Result := [];

  if( ( SourcePin <> NIL ) and ( Assigned(SubmitFunction) ) ) then
    Result := SubmitFunction ( SourcePin, Operation, Data, [nsLastIteration] );

end;
//---------------------------------------------------------------------------
function  TOWSinkPin.DoTryLink( Root : TComponent; SaveForm : Boolean ) : Boolean;
var
  Values        : TStringList;
  I             : Integer;
  RootName      : String;
  MyRoot        : TComponent;

begin
  if ( FInTryLink ) then
    begin
    Result := False;
    Exit;
    end;

  if ( FLinkStorage = '' ) then
    begin
    Result := True;
    Exit;
    end;

  FInTryLink := True;

  Values := TStringList.Create;

  try
    MyRoot := OWGetMainOwnerComponent( Owner );
    if( ( MyRoot <> Root ) and ( Root <> NIL )) then
      begin
      RootName := Root.Name;
      OWGetPinsValueListSingle( Values, Root, Self, '.', True, RootName, SaveForm );
      end

    else
      OWGetPinsValueList( Values, Self, '.', True, SaveForm );

//    FSourcePin := NIL;
    for I := 0 to Values.Count - 1 do
      begin
      if ( FLinkStorage = Values.Strings [ i ] ) then
        begin
        Connect( TOWSourcePin( Values.Objects[ i ] ) );
{
        FSourcePin     := TOWSourcePin( Values.Objects[ i ] );
        SubmitFunction := GetSubmitFunction ( FSourcePin );
        FUpStreamLinkName := GetLinkName ( FSourcePin );
        FSourcePin.IntConnect( TOWSinkPin(Self) );
        FDownStreamLinkName := FSourcePin.GetLinkName ( TOWSinkPin(Self) );
}
        FLinkStorage := '';
        Break;
        end;

      end;

  except
  end;
  Values.Free;
  FInTryLink := False;

  Result := FLinkStorage = '';
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.ClearPending();
begin
  FLinkStorage := '';
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.Connect( OtherPin : TOWPin ) : Boolean;
begin
  OWLinkAwaitsLinking( Owner );

  if ( OtherPin = FSourcePin ) then
    begin
    Result := True;
    Exit
    end;

  if ( OtherPin = NIL ) then
    begin
//    if( FSourcePin <> NIL ) then
    FSourcePin.IntDisconnect(Self);

    FSourcePin := NIL;
    SubmitFunction := NIL;
    Result := True;
    OWNotifyChangePin( Self );
    Exit
    end;

  if ( CanConnectTo ( OtherPin )) then
    begin
    if ( FSourcePin <> NIL ) then
      begin
      FSourcePin.IntDisconnect(Self);
      FSourcePin := NIL;
      SubmitFunction := NIL;
      end;

    FSourcePin := TOWSourcePin( OtherPin );
    SubmitFunction := GetSubmitFunction ( TOWSourcePin( OtherPin ) );
    FUpStreamLinkName := GetLinkName ( FSourcePin );
    FSourcePin.IntConnect(Self);
    FDownStreamLinkName := FSourcePin.GetLinkName ( TOWSinkPin(Self) );

    Result := True;
    FSourcePin.NewConnection( Self, FSourcePin.GetSubmitFunction( Self ));
    OWNotifyChangePin( Self );
    Exit
    end;

  Result := False;
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.Disconnect();
begin
  Connect( NIL );
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.IsCompatible ( const OtherPin : TOWPin ) : Boolean;
var
  SourcePin    : TOWSourcePin;
  I            : Integer;

begin
  if ( OtherPin = NIL ) then
    begin
    Result := True;
    Exit
    end;

  if ( not ( OtherPin is TOWSourcePin )) then
    begin
    Result := False;
    Exit
    end;

  SourcePin := TOWSourcePin( OtherPin );
          
  Result := CheckInterfaces( SourcePin );
  if( not Result ) then
    Result := FIgnoreUpstream;

  if( Result ) then
    begin

    for I := 0 to FDataTypeDependants.Count -1 do
      if( not TOWSourcePin( FDataTypeDependants.Items[ I ] ).ChainCheckInterfaces( Self ) ) then
        begin
        Result := False;
        Exit;
        end
        
    end;

end;
//---------------------------------------------------------------------------
function  TOWSinkPin.IsCompatibleChain ( const OtherPin : TOWPin ) : Boolean;
begin
  Result := IsCompatible( OtherPin );
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.CanConnectTo ( const OtherPin : TOWPin ) : Boolean;
begin
  Result := False;
  if ( OtherPin = NIL ) then
    Exit;

  if ( OtherPin = Self ) then
    Exit;
    
  if ( IsCompatibleChain( OtherPin )) then
    Result := not OtherPin.DependsOn( Self );

end;
//---------------------------------------------------------------------------
function  TOWSinkPin.IsConnected () : Boolean;
begin
  OWLinkAwaitsLinking( Owner );
  Result := ( FSourcePin <> NIL ); 
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.DependsOn ( const OtherPin : TOWPin ) : Boolean; 
begin
  OWLinkAwaitsLinking( Owner );
  if ( OtherPin = FSourcePin ) then
    begin
    Result := True;
    Exit;
    end;

  if ( FSourcePin = NIL ) then
    begin
    Result := False;
    Exit;
    end;

  Result := FSourcePin.DependsOn( OtherPin );
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.Assign(Source: TPersistent);
var
  OtherPin : TOWSinkPin;
  
begin
  if ( Source = NIL ) then
    Connect( NIL ) //Clear ();

  else
    begin
    OtherPin := Source as TOWSinkPin;
    if ( OtherPin <> NIL ) then
      Connect( OtherPin.SourcePin )

    else
      inherited;
      
    end;
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.SetSourcePin ( SourcePin : TOWSourcePin );
begin
  Connect( SourcePin );
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.GetSourcePin () : TOWSourcePin;
begin
  OWLinkAwaitsLinking( Owner );
  Result := FSourcePin;
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.ReadStrData ( Reader : TReader );
begin
//  ShowMessage( Reader.Root.Name );
  TryLinkTo( Reader.Root, Reader.ReadIdent(), True, True );
  OWLinkAwaitsLinking( Reader.Root );
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.TryLinkTo ( Root : TComponent; Soutce : String; TryLater : Boolean; SaveForm : Boolean ) : Boolean;
begin
  Result := True;

  FLinkStorage := Soutce;
  if( Soutce = 'nil' ) then
    begin
    FLinkStorage := '';
    Exit;
    end;
    
  if( FLinkStorage = '' ) then
    begin
    Connect( NIL );
    FLinkStorage := '';
    Exit;
    end;

  if ( not DoTryLink( Root, SaveForm )) then
    begin
    if( TryLater ) then
      AddToAwaitsLinking() // Add to be updated later;
      
    else
      FLinkStorage := '';

    Result := False;
    Exit;
    end;

end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.WriteStrData( Writer : TWriter );
begin
  Writer.WriteIdent ( GetLinkStr( True ) );
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.GetLinkStr( SaveFormat : Boolean ) : String;
var
  AddRoot : Boolean;
           
begin
{           
  if( FLinkStorage <> '' ) then
    begin
    Result := FLinkStorage;
    Exit;
    end;
}   
  AddRoot := False;
  if( SourcePin <> NIL ) then
    AddRoot := OWGetMainOwnerComponent( SourcePin.Owner ) <> OWGetMainOwnerComponent( Owner );
    
  Result := OWValueToString( SourcePin, '.', AddRoot, SaveFormat );
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.AddToAwaitsLinking();
begin
  if( AwaitsLinkingSinkPins.IndexOf( Self ) = -1 ) then
    begin
    AwaitsLinkingSinkPins.Add( Self );
    AllLinked := False;
    end;
    
end;
//---------------------------------------------------------------------------
function  TOWSinkPin.DoWrite() : Boolean;
begin
  Result := ( SourcePin <> NIL ) or ( FLinkStorage <> '' );
end;
//---------------------------------------------------------------------------
procedure TOWSinkPin.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty( 'SourcePin', ReadStrData, WriteStrData, DoWrite());
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
function IsDescendOf ( PropertyObject : TObject; TypeName : String ) : Boolean;
var
  AncestorClass : TClass;
  
begin
  if ( TypeName = PropertyObject.ClassName() ) then
    begin
    Result := True;
    Exit;
    end;

  AncestorClass := PropertyObject.ClassParent();
  while ( AncestorClass <> NIL ) do
    begin
    if ( TypeName = AncestorClass.ClassName() ) then
      begin
      Result := True;
      Exit;
      end;

    AncestorClass := AncestorClass.ClassParent();
    end;

  Result := False;
end;
//---------------------------------------------------------------------------
procedure OWGetClassPropertiesOfTypeName ( AClass : TObject; TypeName : String; AStrings : TStrings; SaveValue : Boolean );
var
  PropList              : PPropList;
  ClassTypeInfo         : PTypeInfo;
  ClassTypeData         : PTypeData;
  i                     : Integer;
  TypeInfo              : TTypeInfo;
  PropertyObject        : TObject;
  PinIndex              : Integer;
//  PersistentClass       : TPersistentClass;
//  Ret   : Boolean;

begin
{ This method retrieves the property names and types for the given object
  and adds that information to the AStrings parameter. }


  ClassTypeInfo := PTypeInfo(AClass.ClassInfo ());
  ClassTypeData := GetTypeData(ClassTypeInfo);

  if ( ClassTypeData.PropCount <> 0 ) then
    begin
    // allocate the memory needed to hold the references to the TPropInfo
    // structures on the number of properties.
    GetMem( PropList, sizeof( PPropInfo ) * ClassTypeData.PropCount );
//    PropList := new PPropInfo [ ClassTypeData->PropCount ];
    try
      // fill PropList with the pointer references to the TPropInfo structures
      GetPropInfos( PTypeInfo( AClass.ClassInfo () ), PropList);
      for i := 0 to ClassTypeData.PropCount - 1 do
        begin
        TypeInfo := PropList[i].PropType^^;
{
        PersistentClass := GetClass( TypeInfo.Name );
        Ret := PersistentClass.InheritsFrom( GetClass( TypeName ));
        if ( Ret ) then
          AStrings.AddObject ( TypeInfo.Name, NIL );
}
        // filter out properties that are events ( method pointer properties)
        if ( TypeInfo.Kind = tkClass ) then
          begin
//function GetObjectProp(Instance: TObject; PropInfo: PPropInfo;
//  MinClass: TClass = nil): TObject; overload;
          PropertyObject := GetObjectProp ( AClass, PropList[i]);
          if ( PropertyObject <> NIL ) then
            if ( IsDescendOf ( PropertyObject, TypeName )) then
              begin
              AStrings.AddObject ( String(PropList[i].Name), PropertyObject );
              end

            else if( PropertyObject is TOWPinList )then
              begin
              for PinIndex := 0 to TOWPinList( PropertyObject ).Count - 1 do
                begin
                if( IsDescendOf ( TOWPinList( PropertyObject ).Pins[ PinIndex ], TypeName )) then
                  begin
                  if( SaveValue ) then
                    AStrings.AddObject ( String(PropList[i].Name) + '._Pin' + IntToStr( PinIndex ), TOWPinList( PropertyObject ).Pins[ PinIndex ] )

                  else
                    AStrings.AddObject ( String(PropList[i].Name) + '.' + TOWPinList( PropertyObject ).Names[ PinIndex ], TOWPinList( PropertyObject ).Pins[ PinIndex ] );
                    
                  end;
                end;
                
              end;


          end;
        end;
    except
      
    end;
    FreeMem( PropList );

    end;

end;
//---------------------------------------------------------------------------
function OWValueToString( StreamPin : TOWPin; Linking : String; AddRoot : Boolean; SaveValue : Boolean ) : String;
var
  j       : Integer;
  IntList : TStringList;
  LinkPath : String;

begin
  if ( StreamPin = NIL ) then
    begin
    Result := '(Disconnected)';
    Exit;
    end;

  IntList := TStringList.Create;

  OWGetClassPropertiesOfTypeName ( StreamPin.Owner, 'TOWPin', IntList, SaveValue );
  for j := 0 to IntList.Count - 1 do
    if ( IntList.Objects [ j ] = StreamPin ) then
      begin
      LinkPath := StreamPin.Owner.Name + Linking + IntList.Strings [ j ];
      if( AddRoot ) then
        LinkPath := OWGetMainOwnerComponent( StreamPin.Owner ).Name + '.' + LinkPath;
         
      break;
      end;

  IntList.Free;

  Result := LinkPath;
end;           
//---------------------------------------------------------------------------
procedure OWGetPinsValueListSingle( List : TStrings; OwnerComponent : TComponent;  StreamPin : TOWPin; Link : String; FullList : Boolean; RootName : String; SaveValue : Boolean );
begin
  OwnerComponent := OWGetMainOwnerComponent ( OwnerComponent );
  OWGetPinsValueListSingleRoot( List, OwnerComponent, StreamPin, Link, FullList, RootName, SaveValue );
end;
//---------------------------------------------------------------------------
procedure OWGetPinsValueListSingleRoot( List : TStrings; OwnerComponent : TComponent;  StreamPin : TOWPin; Link : String; FullList : Boolean; RootName : String; SaveValue : Boolean );
var
  IntList       : TStringList;
  i, j          : Integer;
  Component     : TComponent;
  CanConnect    : Boolean;

begin      
  IntList := TStringList.Create;
  for i := 0 to OwnerComponent.ComponentCount - 1 do
    begin
    Component := OwnerComponent.Components [ i ];
    IntList.Clear ();
    OWGetClassPropertiesOfTypeName ( Component, 'TOWPin', IntList, SaveValue );
    for j := 0 to IntList.Count - 1 do
      begin
      if ( FullList ) then
        CanConnect := StreamPin.IsCompatible ( TOWPin( IntList.Objects [ j ] ))

      else
        CanConnect := StreamPin.CanConnectTo ( TOWPin( IntList.Objects [ j ] ));

      if ( CanConnect ) then
        if( RootName <> '' ) then
          List.AddObject ( RootName + Link + Component.Name + Link + IntList.Strings [ j ], IntList.Objects [ j ] )

        else
          List.AddObject ( Component.Name + Link + IntList.Strings [ j ], IntList.Objects [ j ] );
        
      end;               

    if ( Component.ComponentCount <> 0 ) then
      OWGetPinsValueListSingleRoot( List, Component, StreamPin, Link, FullList, RootName, SaveValue );
      
    end;
                       
  IntList.Free;
end;

//---------------------------------------------------------------------------
{
type TOWModulesColection = class(TStringList)
public
  procedure GetModules( const FileName, UnitName, FormName, DesignClass: string; CoClasses: TStrings );

end;

procedure TOWModulesColection.GetModules( const FileName, UnitName, FormName, DesignClass: string; CoClasses: TStrings );
begin
  if( FormName <> '' ) then
    Add( FormName );

end;
}

procedure OWGetPinsValueList( List : TStrings; StreamPin : TOWPin; Link : String; FullList : Boolean; SaveValue : Boolean );
var
  MainComponent : TComponent;
//  Forms : TOWModulesColection;
//  Form : TCustomForm;

begin
//  Forms := TOWModulesColection.Create;
//  MainComponent := GetMainOwnerComponent ( StreamPin.Owner );
  
//  if( Assigned( Designer )) then
//    Designer.GetProjectModules( Forms.GetModules );
    
  MainComponent := OWGetMainOwnerComponent ( StreamPin.Owner );
  OWGetPinsValueListSingleRoot( List, MainComponent, StreamPin, Link, FullList, '', SaveValue );

//  Forms.Free;
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
procedure TPinType.AddType ( ID : TGUID; SubmitFunction : TOWSubmitFunction );
var
  Entry : TPinTypeEntry;

begin
  Entry.ID := ID;
  Entry.SubmitFunction := SubmitFunction;
  SetLength( Types, Length( Types ) + 1 );
  Types[ Length( Types ) - 1 ] := Entry;
end;

//---------------------------------------------------------------------------

function TPinType.GetSubmitFunc( Index : Integer )     : TOWSubmitFunction;
begin
  Result := Types[ Index ].SubmitFunction;
end;

//---------------------------------------------------------------------------
function TPinType.GetID( Index : Integer ) : TGUID;
begin
  Result := Types[ Index ].ID;
end;

//---------------------------------------------------------------------------

function TPinType.GetLength() : Integer;
begin
  Result := Length( Types );
end;

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
{
constructor TOWPinList.Create();
begin
  inherited;
  PinsList := TStringList.Create;
  PinsOwner := False;
end;
}
//---------------------------------------------------------------------------

constructor TOWPinList.Create( IsPinsOwner : Boolean );
begin
  inherited Create;
  FPinsList := TStringList.Create;
  FPinsOwner := IsPinsOwner;
end;

//---------------------------------------------------------------------------
destructor TOWPinList.Destroy;
var
  I : Integer;

begin
  OWLinkAwaitsLinking( NIL );

  Clear();
{
  if( FPinsOwner ) then
    for I := 0 to Count - 1 do
      Pins[ I ].Free();
}

  FPinsList.Free;
  if( FLinkStorageSink <> NIL ) then
    FLinkStorageSink.Free();

  if( FLinkStorageSource <> NIL ) then
    begin
    for I := 0 to FLinkStorageSource.Count - 1 do
        TStringList( FLinkStorageSource.Items[ I ] ).Free();

    FLinkStorageSource.Free();
    end;

  inherited;
end;
//---------------------------------------------------------------------------
procedure TOWPinList.AfterConstruction();
begin
  inherited;
  OWNotifyAddPinList( Self );
end;
//---------------------------------------------------------------------------
procedure TOWPinList.BeforeDestruction();
begin
  OWNotifyRemovePinList( Self );
  inherited;
end;
//---------------------------------------------------------------------------
procedure TOWPinList.Clear();
var
  I : Integer;

begin
  if( FPinsOwner ) then
    for I := 0 to Count - 1 do
      Pins[ I ].Free();

  FPinsList.Clear();
  
  LastIndicatedCount := -1;
  
  OWNotifyChangePinList( Self );
end;
//---------------------------------------------------------------------------
function TOWPinList.GetCount() : Integer;
begin
  Result := FPinsList.Count;
end;
//---------------------------------------------------------------------------
procedure TOWPinList.SetCount( ACount : Integer );
begin
end;
//---------------------------------------------------------------------------
procedure TOWPinList.ClearPending();
begin
  if( FLinkStorageSink <> NIL ) then
    FLinkStorageSink.Free();

  FLinkStorageSink := NIL;

  if( FLinkStorageSource <> NIL ) then
    FLinkStorageSource.Free();

  FLinkStorageSource := NIL;
end;
//---------------------------------------------------------------------------
function TOWPinList.GetPin( Item : Integer ) : TOWPin;
begin
  Result := TOWPin( FPinsList.Objects[ Item ] );  
end;
//---------------------------------------------------------------------------
function TOWPinList.GetName( Item : Integer ) : String;
begin
  Result := FPinsList.Strings[ Item ];
  if( Result = '' )then
    Result := 'Pin [' + IntToStr( Item ) + ']';
end;
//---------------------------------------------------------------------------
procedure TOWPinList.SetName( Item : Integer; Name : String );
begin
  if( Item >= FPinsList.Count ) then
    Exit;

  if( Name = 'Pin [' + IntToStr( Item ) + ']' ) then
    Name := '';

  if( FPinsList.Strings[ Item ] <> Name ) then
    begin
    FPinsList.Strings[ Item ] := Name;
    LastIndicatedCount := -1;
    end;
    
end;
//---------------------------------------------------------------------------
function  TOWPinList.GetLinkStrInt( Offset : String ) : String;
var
  I     : Integer;
  J     : Integer;
  Sinks : TStringList;

begin
  Result := '';

  if( FLinkStorageSink <> NIL ) then
    begin
    for I := 0 to FLinkStorageSink.Count - 1 do
      Result := Result + Offset + FLinkStorageSink.Strings[ I ] + #13;

    end;

  if( FLinkStorageSource <> NIL ) then
    begin
    for I := 0 to FLinkStorageSource.Count - 1 do
      begin
      Result := Result + #13;
      Sinks := TStringList( FLinkStorageSource.Items[ I ] );
      if( Sinks <> NIL ) then
        for J := 0 to Sinks.Count - 1 do
          if( Sinks.Strings[ J ] <> '' ) then
            Result := Result + Offset + Sinks.Strings[ J ] + #13;
        
      end;

    end;

end;
//---------------------------------------------------------------------------
procedure TOWPinList.Add( Item : TOWPin );
begin
  FPinsList.AddObject( '', Item ); 
  OWNotifyChangePinList( Self );
end;
//---------------------------------------------------------------------------
procedure TOWPinList.AddNamed( Item : TOWPin; Name : String );
begin
  FPinsList.AddObject( Name, Item ); 
  OWNotifyChangePinList( Self );
end;
//---------------------------------------------------------------------------
procedure TOWPinList.Delete( Index : Integer );
begin
  if( FPinsOwner ) then
    TOWPin( FPinsList.Objects[ Index ] ).Free();

  FPinsList.Delete( Index );
  OWNotifyChangePinList( Self );
end;
//---------------------------------------------------------------------------
procedure TOWPinList.Remove( Item : TOWPin );
var
  I : Integer;

begin
  for I:= 0 to FPinsList.Count - 1 do
    if( FPinsList.Objects[ I ] = Item )then
      begin
      if( FPinsOwner ) then
        Item.Free();

      FPinsList.Delete( I );
      end;
      
  OWNotifyChangePinList( Self );
end;
//---------------------------------------------------------------------------
procedure TOWPinList.ReadStrData ( Reader : TReader );
var
  Ident         : String;
  SourceLinks   : TStringList;

begin
  if( FLinkStorageSink = NIL ) then
    FLinkStorageSink := TStringList.Create;
    
  if( FLinkStorageSource = NIL ) then
    FLinkStorageSource := TList.Create;

  Reader.ReadListBegin();
  while not Reader.EndOfList do
    begin
    Ident := Reader.ReadIdent();
    SourceLinks := TStringList.Create();
    if( Ident <> 'nil' ) then
      begin
      if( Ident = 'Sinks' ) then
        begin
        Reader.ReadListBegin();
        while not Reader.EndOfList do
          begin
          Ident := Reader.ReadIdent();
          SourceLinks.Add( Ident ); 
          end;
          
        Reader.ReadListEnd();
        end
        
      else
        FLinkStorageSink.Add( Ident );
        
      end

    else
      FLinkStorageSink.Add( '' );
      FLinkStorageSource.Add( SourceLinks );

    end;

  Reader.ReadListEnd();

  if( AwaitsLinkingLists.IndexOf( Self ) = -1 ) then
    begin
    AwaitsLinkingLists.Add( Self );
    AllLinked := False;
    end;

end;
//---------------------------------------------------------------------------
procedure TOWPinList.WriteStrData( Writer : TWriter );
var
  I      : Integer;

begin
  OWLinkAwaitsLinking( NIL );
  Writer.WriteListBegin();
  for I := 0 to Count - 1 do
    begin
    if( Pins[ I ] is TOWSinkPin ) then
      begin
      if( TOWSinkPin( Pins[ I ] ).DoWrite() ) then
        begin
        TOWSinkPin( Pins[ I ] ).WriteStrData( Writer );
        Continue;
        end
      end
      
    else if( Pins[ I ] is TOWSourcePin ) then
      begin
      if( TOWSourcePin( Pins[ I ] ).DoWrite() ) then
        begin
        Writer.WriteIdent( 'Sinks' );
        TOWSourcePin( Pins[ I ] ).WriteStrData( Writer );
        Continue;
        end
      end;

    Writer.WriteIdent( 'nil' );
    end;
    
  Writer.WriteListEnd();

end;
//---------------------------------------------------------------------------
function  TOWPinList.DoWrite() : Boolean;
var
  I : Integer;

begin
  Result := True;
  for I := 0 to Count - 1 do
    if( Pins[ I ].IsConnected() ) then
      Exit;

  Result := False;

end;
//---------------------------------------------------------------------------
procedure TOWPinList.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty( 'Pins', ReadStrData, WriteStrData, DoWrite());
end;
//---------------------------------------------------------------------------
function TOWPinList.DoTryLink(): Boolean;
var
  I       : Integer;
  Counter : Integer;

begin
//  Result := True;

  if( FLinkStorageSink <> NIL ) then
    begin
    Counter := FLinkStorageSink.Count;
    for I := 0 to FLinkStorageSink.Count - 1 do
      if( FLinkStorageSink.Strings[ I ] <> '' ) then
        begin
        if( I < Count ) then
          if( Pins[ I ] is TOWSinkPin ) then
            begin
            TOWSinkPin( Pins[ I ] ).TryLinkTo ( NIL, FLinkStorageSink.Strings[ I ], True, True );
            FLinkStorageSink.Strings[ I ] := '';
            Dec( Counter );
            end;

        end

      else
        Dec( Counter );

    if ( Counter = 0 ) then
      begin
      FLinkStorageSink.Free();
      FLinkStorageSink := NIL;
      end;
    end;

  if( FLinkStorageSource <> NIL ) then
    begin
    Counter := FLinkStorageSource.Count;
    for I := 0 to FLinkStorageSource.Count - 1 do
      if( TStringList( FLinkStorageSource.Items[ I ] ).Count > 0 ) then
        begin
        if( I < Count ) then
          if( Pins[ I ] is TOWSourcePin ) then
            begin
            TOWSourcePin( Pins[ I ] ).TryLinkToList ( NIL, TStringList( FLinkStorageSource.Items[ I ] ) );
            TStringList( FLinkStorageSource.Items[ I ] ).Clear();
            Dec( Counter );
            end;

        end

      else
        Dec( Counter );

    if ( Counter = 0 ) then
      begin
      for I := 0 to FLinkStorageSource.Count - 1 do
        TStringList( FLinkStorageSource.Items[ I ] ).Free();

      FLinkStorageSource.Free();
      FLinkStorageSource := NIL;
      end;
    end;
    
  Result := ( ( FLinkStorageSource = NIL ) and ( FLinkStorageSink = NIL ) );
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
constructor TOWPinListOwner.Create( APinCreateFunction : TOWPinCreateFunction; APinDestroyFunction : TOWPinDestroyFunction );
begin
  inherited Create( True );
  FMinPins := 0;
  FMaxPins := 100;
  
  FPinCreateFunction := APinCreateFunction;
  FPinDestroyFunction := APinDestroyFunction;
end;
//---------------------------------------------------------------------------
constructor TOWPinListOwner.CreateEx( AMinPins : Integer; AMaxPins : Integer; APinCreateFunction : TOWPinCreateFunction; APinDestroyFunction : TOWPinDestroyFunction = NIL );
begin
  inherited Create( True );
  
  FMinPins := AMinPins;
  
  if( FMinPins < 0 ) then
    FMinPins := 0;

  FMaxPins := AMaxPins;

  FPinCreateFunction  := APinCreateFunction;
  FPinDestroyFunction := APinDestroyFunction;
  
  SetCount( FMinPins );
end;
//---------------------------------------------------------------------------
{
destructor TOWPinListOwner.Destroy;
var
  I : Integer;

begin
  for I := 0 to Count - 1 do
    Pins[ I ].Free();

  inherited;
end;
}
//---------------------------------------------------------------------------
procedure TOWPinListOwner.SetCount( ACount : Integer );
var
  Pin : TOWPin;
begin
  inherited;
  if( ACount < FMinPins )then
    ACount := FMinPins;

  if( ACount > FMaxPins )then
    ACount := FMaxPins;

  while( Count > ACount ) do
    begin
    if( Assigned( FPinDestroyFunction )) then
      FPinDestroyFunction( Self );

    Pin := Pins[ Count - 1 ];
    Remove( Pin );
//    Pin.Free;
    end;

  if( Assigned( FPinCreateFunction )) then
    while( Count < ACount ) do
      begin
      Add( FPinCreateFunction( Self ));
      end;

end;
//---------------------------------------------------------------------------
procedure TOWPinListOwner.ReadStrData ( Reader : TReader );
var
  Ident         : String;
  Pin           : TOWPin;
  DecCounter    : Integer;
  Counter       : Integer;

begin
  DecCounter := Count;
  Counter    := 0;
  Reader.ReadListBegin();
  while not Reader.EndOfList do
    begin
    Ident := Reader.ReadIdent();
    if( Assigned( FPinCreateFunction )) then
      begin
      if( DecCounter = 0 ) then
        begin
        Pin := FPinCreateFunction( Self );
        Add( Pin );
        end

      else
        begin
        Pin := Pins[ Counter ];
        Dec( DecCounter );
        Inc( Counter );
        end;

      if( Ident <> 'nil' )then
        if( Pin is TOWSinkPin )then
          TOWSinkPin( Pin ).TryLinkTo( Reader.Root, Ident, True, True )

        else if( Pin is TOWSourcePin )then
          TOWSourcePin( Pin ).TryLinkTo( Reader.Root, Ident, True, True )

      end;

    end;

  Reader.ReadListEnd();

end;
//---------------------------------------------------------------------------
function  TOWPinListOwner.DoWrite() : Boolean;
begin
  Result := True;
  Exit;
end;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
procedure Register;
begin
end;

var
  Loaded : Boolean;
  
initialization
  IsInLinking := False;
  AwaitsLinkingSourcePins       := TList.Create();
  AwaitsLinkingSinkPins         := TList.Create();
  AwaitsLinkingLists            := TList.Create();
  Loaded := True;
  AllLinked := False;

  NotifyList                    := TList.Create();

finalization
  if( Loaded ) then
    begin
    NotifyList.Free();
    Loaded := False;
    AwaitsLinkingLists.Free();
    AwaitsLinkingSinkPins.Free();
    AwaitsLinkingSourcePins.Free();
    end;
//---------------------------------------------------------------------------
end.

